#include <cstdio>

using namespace std;

#include "ctinsert.hh"
#include "ctcset.hh"
#include "config.hh"
#include "rom.hh"

namespace
{
    void GeneratePatch(ROM &ROM, unsigned offset, const char *fn)
    {
        unsigned MaxHunkSize = GetConf("patch", "maxhunksize");

        /*
        {FILE *fp = fopen("chrono-uncompressed.smc", "rb");
        if(fp){fseek(fp,512,SEEK_SET);fread(&ROM[0], 1, ROM.size(), fp);fclose(fp);}}
        */
        
        /* Now write the patch */
        FILE *fp = fopen(fn, "wb");
        if(!fp)
        {
            perror(fn);
            return;
        }
        fwrite("PATCH", 1, 5, fp);

        /* Format:   24bit offset, 16-bit size, then data; repeat */
        for(unsigned a=0; a<ROM.size(); ++a)
        {
            if(!ROM.touched(a))continue;
            
            putc(((a+offset)>>16)&255, fp);
            putc(((a+offset)>> 8)&255, fp);
            putc(((a+offset)    )&255, fp);
            
            unsigned offs=a, c=0;
            while(a < ROM.size() && ROM.touched(a) && c < MaxHunkSize)
                ++c, ++a;
            
            // Size is "c" in both.
            putc((c>> 8)&255, fp);
            putc((c    )&255, fp);
            fwrite(&ROM[offs], 1, c, fp);
        }
        fwrite("EOF",   1, 3, fp);
        fclose(fp);
        fprintf(stderr, "Created %s\n", fn);
    }
    void GeneratePatches(ROM &ROM)
    {
        GeneratePatch(ROM, 0,   WstrToAsc(GetConf("patch", "patchfn_nohdr")).c_str());
        GeneratePatch(ROM, 512, WstrToAsc(GetConf("patch", "patchfn_hdr")).c_str());
    }
}

const string DispString(const string &s)
{
    static wstringOut conv(getcharset());

    string result;
    for(unsigned a=0; a<s.size(); ++a)
    {
        unsigned char c = s[a];
        ucs4 u = getucs4(c);
        if(u != ilseq)
        {
            result += conv.putc(u);
            continue;
        }
        switch(c)
        {
            case 0x05: result += conv.puts(AscToWstr("[nl]")); break;
            case 0x06: result += conv.puts(AscToWstr("[nl3]")); break;
            case 0x0B: result += conv.puts(AscToWstr("[pause]")); break;
            case 0x0C: result += conv.puts(AscToWstr("[pause3]")); break;
            case 0x00: result += conv.puts(AscToWstr("[end]")); break; // Uuh?
            case 0xEE: result += conv.puts(AscToWstr("[musicsymbol]")); break;
            case 0xF1: result += conv.puts(AscToWstr("...")); break;
            default:
            {
                char Buf[8];
                sprintf(Buf, "[%02X]", c);
                result += conv.puts(AscToWstr(Buf));
            }
        }
    }
    return result;
}

void stringoffsmap::GenerateNeederList()
{
    neederlist.clear();
    
    for(unsigned parasitenum=0; parasitenum < size(); ++parasitenum)
    {
        const string &parasite = (*this)[parasitenum].str;
        for(unsigned hostnum = 0; hostnum < size(); ++hostnum)
        {
            // Can't depend on self
            if(hostnum == parasitenum) { Continue: continue; }
            // If the host depends on this "parasite", skip it
            neederlist_t::iterator i;
            for(unsigned tmp=hostnum;;)
            {
                i = neederlist.find(tmp);
                if(i == neederlist.end()) break;
                tmp = i->second;
                // Host depends on "parasite", skip this
                if(tmp == parasitenum) { goto Continue; }
            }
            
            const string &host = (*this)[hostnum].str;
            if(host.size() < parasite.size())
                continue;
            
            unsigned extralen = host.size()-parasite.size();
            if(parasite == host.substr(extralen))
            {
                for(;;)
                {
                    i = neederlist.find(hostnum);
                    if(i == neederlist.end()) break;
                    // Our "host" depends on someone else.
                    // Take his host instead.
                    hostnum = i->second;
                }
                
                neederlist[parasitenum] = hostnum;
                
                // Now if there are parasites referring to this one
                for(i=neederlist.begin(); i!=neederlist.end(); ++i)
                    if(i->second == parasitenum)
                    {
                        // rerefer them to this one's host
                        i->second = hostnum;
                    }

                break;
            }
        }
    }
    
    /* Nyt mapissa on listattu, kuka tarvitsee ketkin. */

#if 0
    typedef map<string, vector<unsigned> > needertmp_t;
    needertmp_t needertmp;
    for(neederlist_t::const_iterator j = neederlist.begin(); j != neederlist.end(); ++j)
    {
        needertmp[(*this)[j->first].str].push_back(j->first);
    }
    for(needertmp_t::const_iterator j = needertmp.begin(); j != needertmp.end(); ++j)
    {
        unsigned hostnum = 0;
        const vector<unsigned> &tmp = j->second;

        unsigned c=0;
        fprintf(stderr, "String%s", tmp.size()==1 ? "" : "s");
        for(unsigned a=0; a<tmp.size(); ++a)
        {
            hostnum = neederlist[tmp[a]];
            if(++c > 15) { fprintf(stderr, "\n   "); c=0; }
            fprintf(stderr, " %u", tmp[a]);
        }
        
        fprintf(stderr, " (%s) depend%s on string %u(%s)\n",
            DispString(j->first).c_str(),
            tmp.size()==1 ? "s" : "", hostnum,
            DispString((*this)[hostnum].str).c_str());
    }
#endif
}

const set<unsigned> insertor::GetZStringPageList() const
{
    set<unsigned> result;
    for(stringmap::const_iterator i=strings.begin(); i!=strings.end(); ++i)
    {
        unsigned page = i->first >> 16;
        switch(i->second.type)
        {
            case stringdata::fixed:
                // This is not a pointer
                continue;
            case stringdata::zptr8:
            case stringdata::zptr12:
                // These are ok
                break;
            // If we omitted something, compiler should warn
        }
        result.insert(page);
    }
    return result;
}

const stringoffsmap insertor::GetZStringList(unsigned pagenum) const
{
    stringoffsmap result;
    for(stringmap::const_iterator i=strings.begin(); i!=strings.end(); ++i)
    {
        unsigned page = i->first >> 16;
        if(page != pagenum) continue;
        switch(i->second.type)
        {
            case stringdata::fixed:
                // This is not a pointer
                continue;
            case stringdata::zptr8:
            case stringdata::zptr12:
                // These are ok
                break;
            // If we omitted something, compiler should warn
        }
        stringoffsdata tmp;
        tmp.str  = i->second.str;
        tmp.offs = i->first;
        result.push_back(tmp);
    }
    return result;
}


int main(void)
{
    fprintf(stderr,
        "Chrono Trigger script insertor version "VERSION"\n"
        "Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)\n");
    
    insertor *ins = new insertor;
    
    string font8fn  = WstrToAsc(GetConf("font",   "font8fn"));
    string font8vfn = WstrToAsc(GetConf("font",   "font8vfn"));
    string font12fn = WstrToAsc(GetConf("font",   "font12fn"));
    string scriptfn = WstrToAsc(GetConf("readin", "scriptfn"));
    
    // Font loading must happen before script loading,
    // or script won't be properly paragraph-wrapped.
    ins->LoadFont8(font8fn);
    ins->LoadFont8v(font8vfn);
    ins->LoadFont12(font12fn);
    
    FILE *fp = fopen(scriptfn.c_str(), "rt");
    ins->LoadFile(fp);
    fclose(fp);
    
    ins->DictionaryCompress();
    
    /////////ins->GenerateCode();
    
    ins->freespace.Report();

    fprintf(stderr, "Creating a virtual ROM...\n");
    ROM ROM(4194304);
    
    ins->PatchROM(ROM);

    ins->freespace.Report();
    
    fprintf(stderr, "Unallocating insertor data...\n");
    delete ins; ins = NULL;
    
    GeneratePatches(ROM);
    
    return 0;
}
